/* zklEditLine.c : Front end for BSD editline and history libraries
 * Bog simple, I basically copied the example from
 *    http://www.cs.utah.edu/~bigler/code/libedit.html
 *    man el_init
 * 
 * Libraries you will need: libedit2, libedit-dev, libncurses
 * 
 * Copyright (c) 2006, 2007, 2008, 2013 Craig Durland craigd@zenkinetic.com
 * License: The zlib/libpng License,
 * 	http://www.opensource.org/licenses/zlib-license.php
 *
 * libedit is BSDized code.
 * 
 * To enter an ESC, use C-v ESC
 * 
 * History file: ~/.zklHistory
 * If you want a directory local history file, "touch .zklHistory"
 * And yes, two instances of zkl, last one to exit wins (if history is in
 * the same place).
 */

#define HISTORY_FILE	   "~/.zklHistory"
#define LOCAL_HISTORY_FILE ".zklHistory"

#include "zklObject.h"
#include "zklList.h"
#include "zklString.h"
#include "zklUtil.h"
#include "zklImports.h"

#include <string.h>
#include <histedit.h>
#include <curses.h>
//#include <errno.h>

static ZKL_Object EditLineObject;
static CAtomicInt refcount;	// I'll only create ONE EditLine Instance

typedef struct
{
   Instance	instance;	// Inherit from Instance
   int		live;
   EditLine    *editor;
   History     *history;
   HistEvent	ev;
} LineEditor;


static char *pooh;	// Argh! Not thread safe
char *prompt(EditLine *el)
{
   return pooh;
}

/* ******************************************************************** */
/* ***************************** Methods ****************************** */
/* ******************************************************************** */

    // .ask(prompt) --> String
static Instance *le_ask(LineEditor *self,pArglist arglist,pVM vm)
{
   extern void vmProcessInterrupt(pVM);	// vm.c

   int   count;
   char *line;
   char *prompt = arglistTryToGetString(arglist,0,".ask",vm);

   if (!self->live)
      vmThrow(vm,E_ASSERTION_ERROR,"LineEditor.ask: Not initialized! Create an instance.");

   pooh = prompt ? prompt : "";
   vmMightStall(vm,1);
	// count is the number of characters read. line has a tailing \n
      line = (char *)el_gets(self->editor, &count);
   vmMightStall(vm,0);

	// In order to use our history we have to explicitly add commands
	// to the history
   if (count > 1)	// ignore empty line
   {
      history(self->history, &self->ev, H_ENTER, line);
      line[count - 1] = '\0';	// nuke trailing \n
   }
   else if (count == 0) line = "";	// control-D == EoL
   else	if (count < 0)	// error
   {
if(count)printf(">>>ERROR  %d %d %d\n",count,*line,*vmInterruptEvent(vm));
      if (*vmInterruptEvent(vm))
	 vmProcessInterrupt(vm);	// longjmp()s
      count = 0;
   }
   return stringCreate(line,I_OWNED,vm);
}

    // .readKey() --> String, ^D works, -->"^D" == 4
static Instance *le_getc(LineEditor *self,pArglist arglist,pVM vm)
{
   int  count, n=1;
   char c;

   if (!self->live)
      vmThrow(vm,E_ASSERTION_ERROR,"LineEditor.ask: Not initialized! Create an instance.");

   vmMightStall(vm,1);
	// count is the number of characters read. line has a tailing \n
      count = el_getc(self->editor, &c);
   vmMightStall(vm,0);

   if (count < 1) n = 0;
   return stringCreate2(&c,n,vm);
}

static int _saveName(char *fileName, void *fxp)
{
   strcpy((char *)fxp,fileName);
//printf("-->%s\n",(char *)fxp);
   return 2;	// stop searching
}
static int findHistorFileName(int forClose, char *buf)
{
   int flags = (FXP_NO_DIRS | FXP_ONLY_ONE);

   	// use .zklHistory only if it exists
   	// yeah, yeah, I should just fstat() it
   if (2 == fxpand(LOCAL_HISTORY_FILE,flags,_saveName,buf)) return 2;
   if (forClose) flags |= FXP_DF;
   	// otherwise, use ~/.zklHistory
   if (2 == fxpand(HISTORY_FILE,flags,_saveName,buf)) return 2;
   return 0;
}

unsigned char controlC(EditLine *e,int ch)
{
   printf("CONTROLC\n");
   return CC_NORM;
}

    // .create()
    // Initialize a EditLine state to use our prompt function and
    // emacs style editing.
static Instance *le_create(Instance *x,pArglist arglist,pVM vm)
{
   LineEditor  *lineEditor;
   EditLine    *el;
   History     *h;

   if (!SPIN_LOCK_TRY_TO_ACQUIRE(&refcount))
      vmThrow(vm,E_ASSERTION_ERROR,"EditLine.create: already in use");

   lineEditor = (LineEditor *)
		instanceAllocate(sizeof(LineEditor),&EditLineObject,1,vm);

   el = el_init("zkl",stdin,stdout,stderr);
   el_set(el, EL_PROMPT, &prompt);
   el_set(el, EL_EDITOR, "emacs");
//el_set(el, EL_ADDFN, CC_NORM,0,CONTROLC);
   lineEditor->editor = el;

   h = history_init();
   if (h)
   {
      char buf[2000];
      HistEvent ev;
	// Set the size of the history
      history(h, &lineEditor->ev, H_SETSIZE, 800);
      history(h, &lineEditor->ev, H_SETUNIQUE,1);  // ignore duplicate lines
      if (findHistorFileName(0,buf)) history(h,&ev,H_LOAD,buf);

     	// This sets up the call back functions for history functionality
      el_set(el, EL_HIST, history, h);
   }
   lineEditor->history = h;
   lineEditor->live    = 1;
   return (Instance *)lineEditor;
}

static int editLineClose(Instance *_self)	// also GC routine
{
   LineEditor *self = (LineEditor *)_self;
   if (self->live) 
   {
      if (self->history)
      {
	 char buf[2000];
	 HistEvent ev;

	 if (2 == findHistorFileName(1,buf))
	     history(self->history,&ev,H_SAVE,buf);
	 history_end(self->history);	// Clean up memory
      }
      el_end(self->editor);
      CAI_DEC(&refcount);
      self->live = 0;
   }
   return 1;
}

    // .close()
static Instance *le_close(Instance *self,pArglist arglist,pVM vm)
   { editLineClose(self); return self; }

    // .toBool()
static Instance *le_toBool(LineEditor *self,pArglist arglist,pVM vm)
   { return boolCreate(self->live); }

static MethodTable methods[] =
{
   "create",	(pMethod)le_create,
   "close",	(pMethod)le_close,
   "ask",	(pMethod)le_ask,
   "readKey",	(pMethod)le_getc,
   "read",	(pMethod)le_ask,	// Stream
   "readln",	(pMethod)le_ask,	// Stream
   "toBool",	(pMethod)le_toBool,
   0,		0
};


/* ******************************************************************** */
/* **************************** Properties **************************** */
/* ******************************************************************** */

    // -->Bool
static Instance *keyPressed(Instance *self,pVM vm)
{
#if 0	// can't find the magic dust
    #if 0
      fd_set rfds;
      int    s;
      struct timeval tv = { 1,0 };

      FD_ZERO(&rfds); FD_SET(0,&rfds);
      s = select(1,&rfds,0,0,&tv);
printf("-->%d\n",s);
      if (s > 0) return BoolTrue;
    #else	// curses
      int s = getch();
printf("-->%d\n",s);
      if (s != ERR) { ungetch(s); return BoolTrue; }
    #endif
#endif
   return BoolFalse;
}

    // .version
static Instance *version(Instance *self, pVM vm)
   { return intCreate(1,vm); }	// 0x2020

    // .authors
static Instance *authors(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,"Craig Durland","http://zenkinetic.com/",
          (char *)0); 
}

    // .libraries
static Instance *libraries(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,
      "libedit","",
      (char *)0);
}

static PropertyTable properties[] =
{
   "version",		(pProperty)version,
   "authors",		(pProperty)authors,
   "libraries",		(pProperty)libraries,
   "keyPressed",	(pProperty)keyPressed,
   0,			0
};


//static LineEditor editLineInstance;	// a place holder

DllExport void *construct(void *vm)
{
   if (!vm) return (void *)ZKLX_PROTOCOL;	// handshake
   	// If this is reloaded, nothing happens except 
	// construct() is called again so don't reinitialize
   if (!EditLineObject.name)	// static items are zero
   {
      constructObject(&EditLineObject,NativeType, methods,properties,0,vm);
      EditLineObject.name      = "EditLine";
      EditLineObject.vaultPath = "Console.EditLine";
//      EditLineObject.isize     = sizeof(editLineInstance);
      EditLineObject.isize     = sizeof(LineEditor);
      EditLineObject.freeMe    = editLineClose;

      CAI_INIT(&refcount); 
//      editLineInstance.live = 0; 
//      instanceInit((Instance *)&editLineInstance,&EditLineObject,I_UNTOUCHABLE);
      return le_create(Zero,NoArglist,vm);
   }
//   return &editLineInstance;
//   return Zero;	// Import() doesn't like Void
   return le_create(Zero,NoArglist,vm);	// error on reload
}
